"""
This module contains the service layer for the :mod:`.pipeline` module.
"""

from relevance.service import Service
from relevance.manager import ResourceManager
from relevance.pipeline import Pipeline
from relevance.pipeline import PipelineFactory


# Initialize the service and its dependencies
service = Service.instance('pipeline')
manager = service.data.manager = ResourceManager(Pipeline)
factory = service.data.factory = PipelineFactory(manager)

# Bootstrap
factory.load_all(service.settings, manager)


@service.route('/pipelines', methods=['GET'])
def get_pipelines():
    """
    Get a list of pipelines.

    :Method: GET
    :Path: /pipelines

    :Statuses: `200` on success.
    :Response:
        .. code-block:: json

            ["name1", "name2", "name3"...]
    """
    return service.result(200, list(manager.stack.keys()))


@service.route('/pipelines/<pipeline_name>', methods=['GET'])
def get_pipeline(pipeline_name: str):
    """
    Get information about a pipeline.

    :Method: GET
    :Path: /pipelines/`pipeline_name`
    :param pipeline_name: the name of the pipeline to get information about.

    :Statuses: `200` on success.
    :Response:
        .. code-block:: json

            {
                "components": ["component1", "component2"...],
                "worker": {
                    "is_started": True,
                    "pending_tasks": 42,
                    "type": "relevance.worker.ThreadWorker"
                }
            }
    """
    pipeline = manager.resolve(pipeline_name)
    result = {
        'components': list(pipeline.components.keys()),
        'worker': {
            'is_started': pipeline.worker.is_started,
            'pending_tasks': pipeline.worker.scope.qsize(),
            'type': '{0}.{1}'.format(
                pipeline.worker.__class__.__module__,
                pipeline.worker.__class__.__name__,
            ),
        },
    }
    return service.result(200, result)


@service.route('/pipelines/<pipeline_name>/worker', methods=['GET'])
def get_pipeline_worker(pipeline_name: str):
    """
    Get information about a pipeline's worker.

    :Method: GET
    :Path: /pipelines/`pipeline_name`/worker
    :param pipeline_name: the name of the pipeline to get information about.

    :Statuses: `200` on success.
    :Response:
        .. code-block:: json

            {
                "is_started": True,
                "pending_tasks": 42,
                "type": "relevance.worker.ThreadWorker"
            }
    """
    pipeline = manager.resolve(pipeline_name)
    result = {
        'is_started': pipeline.worker.is_started,
        'pending_tasks': pipeline.worker.scope.qsize(),
        'type': '{0}.{1}'.format(
            pipeline.worker.__class__.__module__,
            pipeline.worker.__class__.__name__,
        ),
    }
    return service.result(200, result)


@service.route('/pipelines/<pipeline_name>/components', methods=['GET'])
def get_pipeline_components(pipeline_name: str):
    """
    Get the list of components for a pipeline.

    :Method: GET
    :Path: /pipelines/`pipeline_name`/components
    :param pipeline_name: the name of the pipeline to get information about.

    :Statuses: `200` on success.
    :Response:
        .. code-block:: json

            ["component1", "component2"...]
    """
    pipeline = manager.resolve(pipeline_name)
    result = list(pipeline.components.keys())
    return service.result(200, result)


@service.route('/pipelines/<pipeline_name>/components/<component_name>', methods=['GET'])
def get_pipeline_component(pipeline_name: str, component_name: str):
    """
    Get information about a specific component in a pipeline.

    :Method: GET
    :Path: /pipelines/`pipeline_name`/components/`component_name`
    :param pipeline_name: the name of the pipeline to get information about.
    :param component_name: the name of the component to get information about.

    :Statuses: `200` on success.
    :Response:
        .. code-block:: json

            {
                "class": "relevance.pipeline.Extractor",
                "target": "mymodule.myfunc",
                "options": {
                    "url": "some param",
                },
                "worker": {
                    "is_started": True,
                    "pending_tasks": 42,
                    "type": "relevance.worker.ThreadWorker"
                }
            }
    """
    pipeline = manager.resolve(pipeline_name)
    component = pipeline.get(component_name)

    result = {
        'class': '{0}.{1}'.format(
            component.__class__.__module__,
            component.__class__.__name__,
        ),
        'target': '{0}.{1}'.format(
            component.target.__module__,
            component.target.__name__,
        ),
        'options': component.options,
        'worker': {
            'is_started': component.worker.is_started,
            'pending_tasks': component.worker.scope.qsize(),
            'type': '{0}.{1}'.format(
                component.worker.__class__.__module__,
                component.worker.__class__.__name__,
            ),
        },
    }
    return service.result(200, result)


@service.route('/pipelines/<pipeline_name>/worker', methods=['PUT'])
def start_pipeline_worker(pipeline_name: str):
    """
    Start a pipeline worker.

    :Method: PUT
    :Path: /pipelines/`pipeline_name`/worker
    :param pipeline_name: the name of the pipeline to start.

    :Statuses: 202 on success.
    :Response: same as :func:`get_pipeline`.
    """
    pipeline = manager.resolve(pipeline_name)
    pipeline.start()

    result = {
        'is_started': pipeline.worker.is_started,
        'pending_tasks': pipeline.worker.scope.qsize(),
        'type': '{0}.{1}'.format(
            pipeline.worker.__class__.__module__,
            pipeline.worker.__class__.__name__,
        ),
    }
    return service.result(202, result)


@service.route('/pipelines/<pipeline_name>/worker', methods=['DELETE'])
def stop_pipeline_worker(pipeline_name: str):
    """
    Stop a pipeline worker.

    :Method: DELETE
    :Path: /pipelines/`pipeline_name`/worker
    :param pipeline_name: the name of the pipeline to stop.

    :Statuses: 202 on success.
    :Response: same as :func:`get_pipeline`.
    """
    pipeline = manager.resolve(pipeline_name)
    pipeline.stop()

    result = {
        'is_started': pipeline.worker.is_started,
        'pending_tasks': pipeline.worker.scope.qsize(),
        'type': '{0}.{1}'.format(
            pipeline.worker.__class__.__module__,
            pipeline.worker.__class__.__name__,
        ),
    }
    return service.result(202, result)


if __name__ == '__main__':
    import os
    service.run(port=int(os.getenv('PORT', 55346)), debug=True)



"""
The mapping module provides an abstract interface for managing index schema mappings across
different search engine implementations.
"""

import typing


class Field(object):
    """
    This is the base class used to represent a mapped field.
    """
    def __init__(self, name: str, data_type: type, stored: bool=True, **options):
        """
        :param name: the name of the field in the index.
        :param data_type: the data type for the field. Python data types (e.g.:
            :class:`str`, :class:`int`, etc.) should be supplied here.
        :param stored: a flag determining if the field should only be indexed or its
            value stored as well. Some engine implementations may ignore this flag.
        :param **options: additional options to store for the mapping, typically engine
            specific, such as unsigned flags, maximum length, etc.
        """
        self.name = name
        self.data_type = data_type
        self.stored = stored
        self.options = options

    def to_dict(self) -> typing.Dict[str, object]:
        """
        Obtain the field mapping as a dictionary object.

        :return: a dictionary containing the field information.
        """
        data_type = '{0}.{1}'.format(
            self.data_type.__module__,
            self.data_type.__name__,
        )

        return {
            'name': self.name,
            'data_type': data_type,
            'stored': self.stored,
            'options': self.options,
        }

    def __repr__(self) -> str:
        """
        Implementation of the :func:`repr` function.
        """
        return '{0}:{1}'.format(self.name, self.data_type.__name__)


class Mapping(object):
    """
    This class is used to represent an index mapping, that is, a collection of
    fields that a particular schema uses.
    """
    def __init__(self, *fields):
        """
        :param *fields: the list of field objects in the mapping.
        """
        self.fields = list(fields)

    def add(self, field: Field):
        """
        Add a field to the mapping.

        :param field: the field object to add.
        """
        self.fields.append(field)

    def to_dict(self) -> typing.Dict[str, dict]:
        """
        Obtain the mapping as a dictionary.

        This method also calls the :meth:`Field.to_dict` method on each field.

        :return: a dictionary object containing the mapping information.
        """
        result = {}
        for x in self.fields:
            result[x.name] = x.to_dict()
        return result

    def __getitem__(self, name: str) -> Field:
        """
        Get a field by its name.

        :param name: the name of the field.
        :return: the field object.
        :raises: :class:`FieldNotFoundError`: when the requested field does not exist.
        """
        for x in self.fields:
            if x.name == name:
                return x
        raise FieldNotFoundError(name)

    def __contains__(self, name: str) -> bool:
        """
        Check if a field is part of the mapping.

        :param name: the field name.
        :return: True if it exists, False otherwise.
        """
        for x in self.fields:
            if x.name == name:
                return True
        return False

    def __iter__(self) -> typing.Iterable:
        """
        Iterate over all the fields.

        :return: an iterator object that iterates over all the mapping's field,
            in order.
        """
        return iter(self.fields)

    def __repr__(self) -> str:
        """
        Implementation of the :func:`repr` function.
        """
        return repr(list(self.fields))


class ObjectField(Field):
    """
    This class provides the interface for mapping an object field. Object fields
    are fields that have nested mappings in them.
    """
    def __init__(self, name: str, mapping: Mapping, *args, **kwargs):
        """
        :param name: the name of the field.
        :param mapping: the mapping object for the field, representing the field's
            sub-fields.
        """
        super().__init__(name, dict, *args, **kwargs)
        self.mapping = mapping

    def to_dict(self) -> typing.Dict[str, object]:
        """
        Implementation of :meth:`Field.to_dict`.
        """
        data = super().to_dict()
        del data['data_type']
        data['fields'] = self.mapping.to_dict()
        return data


class FieldNotFoundError(NameError):
    """
    This exception class is raised when trying to get a field that does not exist in a
    specific mapping.
    """
    pass



"""
Relevance document module.

This module provides the interface for representing documents, search results, and
result sets.
"""

import datetime


class Document(dict):
    """
    Document class.

    This class represents a document that can be searched for or indexed using an engine
    implementation. Documents are dictionaries with a schema name, a type and a unique
    identifier attached.
    """
    def __init__(self, schema: str, doc_type: str, uid: object):
        """
        Initialize the document.

        :param schema: the document schema, typically an index name.
        :param doc_type: the document type, usually a table name or similar.
        :param uid: a unique identifier for this document.
        """
        self.schema = schema
        self.doc_type = doc_type
        self.uid = uid

    def __eq__(self, other: object):
        """
        Overload operator for comparison.

        Documents are considered equal if they have the same content, document type and
        unique identifier.
        """
        return dict(self) == dict(other) and self.uid == other.uid and \
            self.doc_type == other.doc_type


class Result(Document):
    """
    Result class.

    This class provides the interface for handling single search results. Results are
    documents that are returned by a search query. They have the same attributes as
    documents with an extra relevancy score attached to them.
    """
    def __init__(self, *args, score: float):
        """
        Initialize the result.

        :param score: the relevancy score returned from the query result.
        """
        super().__init__(*args)
        self.score = score


class ResultSet(list):
    """
    Result set class.

    This class provides the interface for handling search result lists. Result sets
    are basically list of documents with some additional information such as a
    dictionary of facets and a number of total matches in the schema.
    """
    def __init__(self):
        """
        Initialize the result set.
        """
        self.time_start = datetime.datetime.now()
        self.time_end = None
        self.facets = None
        self.count = 0

    @property
    def time(self) -> float:
        """
        Get the total amount of time the request took.
        """
        return (self.time_end - self.time_start).total_seconds()



"""
This package provides the utilities and interfaces needed to ingest content into a
search backend.
"""

import abc
import pydoc
import typing
import collections

from relevance import loggers
from relevance.worker import Worker
from relevance.worker import WorkerProxy
from relevance.worker import ThreadWorker
from relevance.manager import ResourceManager
from relevance.manager import ResourceExistsError
from relevance.manager import ResourceNotFoundError
from relevance.settings import Settings
from relevance.settings import InvalidSettingValueError


# Logging
logger = loggers.getLogger('relevance.pipeline')


class Component(WorkerProxy, metaclass=abc.ABCMeta):
    """
    The component class is a unit of a pipeline. A pipeline contains multiple components
    that have different roles, such as extracting content, post-processing it and
    publishing it to a queue or search backend. Components are isolated from one another
    and each have their own worker.
    """
    def __init__(self, worker: Worker, target: typing.Callable, **options):
        """
        :param worker: the worker object to wrap the component into.
        :param target: the component's callable. Most components are wrappers for
            existing functions. Whenever the data is ready, the target callable will
            be executed and the component will run.
        :param **options: the component's options. For simple callables, this is usually
            a dictionary of keyword arguments. If the component is a class, it would be
            a dictionary of keyword arguments to pass to the class initializer. For
            custom component implementations, it is component-specific.
        """
        super().__init__(worker)
        self.target = target
        self.options = options

    @abc.abstractmethod
    def __call__(self, pipeline: 'Pipeline', **kwargs) -> object:
        """
        Execute the component.

        This is the main entry point into the component.

        :param pipeline: the pipeline object in which the component is currently being
            executed.
        :param **kwargs: additional arguments to pass to the component at execution time.
        :returns: the result of the component execution.
        """
        kwargs.update(self.options)
        return self.target(**kwargs)


class Extractor(Component):
    """
    An extractor is a component that pulls content from a data source for further
    processing. Per example, it could be an in-memory queue, a message broker, a
    database or a web crawler.

    This implementation simply accepts a callable that will be executed at each run
    in the worker's event loop, pulling content (if available) at each iteration. If
    an item is available, the `item_extracted` signal will be sent to the pipeline with
    the extracted item.
    """
    def __call__(self, pipeline: 'Pipeline') -> object:
        """
        Implementation of :meth:`Component.__call__`.
        """
        item = super().__call__(pipeline)

        if isinstance(item, StopIteration):
            return item

        pipeline.trigger('item_extracted', item=item)
        return item


class Publisher(Component):
    """
    A publisher is a component that pushes content into a backend (such as a different
    queue, a database or a search backend) once it has finished processing.

    This implementation simply accepts a callable that will be executed whenever an
    `item_processed` signal is received. It then executes the callable, and triggers an
    `item_published` signal to the pipeline with the result of the publication.
    """
    def __call__(self, pipeline: 'Pipeline', item: object) -> object:
        """
        Implementation of :meth:`Component.__call__`.
        """
        result = super().__call__(pipeline, item=item)

        if isinstance(item, StopIteration):
            return item

        pipeline.trigger('item_published', item=item, result=result)
        return item


class Processor(Component):
    """
    A processor is a component that modifies the original document before publishing it.

    This implementation accepts a callable that is executed when a `item_extracted` signal
    is received. The callable is executed, the document is modified and an `item_processed`
    signal is broadcasted with the new version of the document.
    """
    def __call__(self, pipeline: 'Pipeline', item: object) -> object:
        """
        Implementation of :meth:`Component.__call__`.
        """
        item = super().__call__(pipeline, item=item)

        if isinstance(item, StopIteration):
            return item

        pipeline.trigger('item_processed', item=item)
        return item


class ProcessorStackMetaclass(abc.ABCMeta):
    """
    This metaclass is used by the processor stack wrapper class to collect the method
    objects to execute in the order they were defined.
    """
    @classmethod
    def __prepare__(metacls: type, name: str, bases: tuple, **kwds):
        """
        Prepare the class namespace.
        """
        return collections.OrderedDict()

    def __new__(cls: type, name: str, bases: tuple, namespace: dict, **kwds):
        """
        Create a new instance of the class.
        """
        result = type.__new__(cls, name, tuple(bases), dict(namespace))

        members = []
        for name, func in namespace.items():
            if hasattr(func, '__call__') and name[0] != '_':
                members.append(func)

        result.members = tuple(members)
        return result


class ProcessorStack(Processor, metaclass=ProcessorStackMetaclass):
    """
    The processor stack is a variant of the :class:`Processor` processor class, that
    allows a class to be executed as a processor rather than a callable. The processor
    stack basically checks all the public methods defined in the class and executes
    them in order. This is useful when a lot of different logic or code paths are
    necessary to process a document, to reduce bloat in the underlying code.
    """
    def __init__(self, worker: Worker):
        """
        :param worker: the worker object to wrap the component into.
        """
        super().__init__(worker, self)

    def __call__(self, pipeline: 'Pipeline', item: object) -> object:
        """
        Implementation of :meth:`Component.__call__`.
        """
        for x in self.members:
            result = x(self, item)

            if isinstance(result, StopIteration):
                return result

        if isinstance(result, StopIteration):
            return result

        pipeline.trigger('item_processed', item=item)
        return item


class Pipeline(WorkerProxy):
    """
    The pipeline class is responsible for managing and executing components. Like
    components, it proxies a worker object.
    """
    def __init__(self, *args, **kwargs):
        """
        :param worker: the worker object to wrap the pipeline into.
        """
        super().__init__(*args, **kwargs)
        self.components = {}

    def add(self, name: str, target: Component):
        """
        Add a component to the pipeline.

        :param name: the name to give the component.
        :param target: the component object.
        :raises: :class:`ResourceExistsError`: if a component of the same
            name already exists.
        """
        if target.worker is not self.worker:
            self.worker.add(target.worker)

        if name in self.components:
            raise ResourceExistsError(name)
        self.components[name] = target

        if isinstance(target, Extractor):
            self.worker.listen('loop_start', target, pipeline=self)
        if isinstance(target, Processor):
            self.worker.listen('item_extracted', target, pipeline=self)
        if isinstance(target, Publisher):
            self.worker.listen('item_processed', target, pipeline=self)

    def get(self, name: str) -> Component:
        """
        Get a component by its name.

        :param name: the name of the component.
        :returns: the component object.
        :raises: :class:`ResourceNotFoundError`: if a component of the specified
            name does not exist.
        """
        try:
            return self.components[name]
        except KeyError:
            raise ResourceNotFoundError(name)


class PipelineFactory(object):
    """
    The pipeline factory is responsible for creating pipeline and component objects
    from settings objects.
    """
    def __init__(self, worker_cls: type=ThreadWorker):
        """
        :param worker_cls: the worker class name.
        """
        self.worker_cls = worker_cls

    def load_component(self, settings: Settings) -> Component:
        """
        Load a component object from settings.

        :param settings: the settings object.
        :returns: the component object.
        :raises: :class:`ImportError`: when the component class, the target or the
            worker class is invalid.
        """
        cls_name = settings.get_type('class', data_type=str)
        target = settings.get('target', None, data_type=str)
        options = settings.get('options', {}, data_type=dict)
        worker_cls_name = settings.get('worker',
                                       '{0}.{1}'.format(
                                           self.worker_cls.__module__,
                                           self.worker_cls.__name__
                                       ),
                                       data_type=str)

        cls = pydoc.locate(cls_name)
        if cls is None or not issubclass(cls, Component):
            raise ImportError(cls_name)

        if target is not None:
            func = pydoc.locate(target)
            if func is None:
                raise ImportError(target)
            target = func

        worker_cls = pydoc.locate(worker_cls_name)
        if worker_cls is None or not issubclass(worker_cls, Worker):
            raise ImportError(worker_cls_name)

        worker = worker_cls()
        if target is not None:
            component = cls(worker, target, **options)
        else:
            component = cls(worker, **options)

        return component

    def load_all_components(self, settings: Settings) -> typing.Dict[str, Component]:
        """
        Load all component objects into a dictionary.

        :param settings: the settings object to load from.
        :returns: a dictionary of name and component object mappings.
        """
        result = {}
        for name, conf in settings.items():
            component = self.load_component(conf)
            result[name] = component
        return result

    def load_all(self, settings: Settings, manager: ResourceManager):
        """
        Load an entire configuration into a manager.

        :param settings: the settings object to load.
        :param manager: the manager object to load it into.
        :raises: :class:`ImportError`: when the pipeline class, or the worker
            classes are invalid.
        """
        worker_cls_name = settings.get('default_worker', None, data_type=str)
        if worker_cls_name is not None:
            worker_cls = pydoc.locate(worker_cls_name)
            if worker_cls is None or not issubclass(worker_cls, Worker):
                raise ImportError(worker_cls_name)
            self.worker_cls = worker_cls

        pipeline_cls_name = settings.get('pipeline_worker', None, data_type=str)
        if pipeline_cls_name is not None:
            pipeline_cls = pydoc.locate(pipeline_cls_name)
            if pipeline_cls is None or not issubclass(pipeline_cls, Worker):
                raise ImportError(pipeline_cls_name)
        else:
            pipeline_cls = self.worker_cls

        conf = settings.get('components', {}, data_type=dict)
        components = self.load_all_components(conf)

        for name, conf in settings.get('pipelines', {}, data_type=dict).items():
            pipeline = manager.new(name, pipeline_cls())

            for key, target in conf.items():
                try:
                    component = components[target]
                except KeyError:
                    raise InvalidSettingValueError(target)

                pipeline.add(key, component)



"""
This module provides the interface definitions for defining and parsing search requests.
It abstracts queries using a specific query language, heavily based on Python syntax.
"""

import enum
import io
import typing
import tokenize

from relevance import loggers


# Logging
logger = loggers.getLogger('relevance.query')


class SortOrder(enum.Enum):
    """
    Operators for sort orders.
    """
    ASC = 'asc'
    DESC = 'desc'


class LogicalOperator(enum.Enum):
    """
    Operators for logical expressions.
    """
    AND = 'and'
    OR = 'or'
    NOT = 'not'


class ComparisonOperator(enum.Enum):
    """
    Operators for comparison expressions.
    """
    EQ = '=='
    NEQ = '!='
    GT = '>'
    LT = '<'
    GTE = '>='
    LTE = '<='


class Query(object):
    """
    This class is used to represent a query. A query is the part of the search
    request that includes terms and facet filters.
    """
    def __init__(self, logic: LogicalOperator=LogicalOperator.AND):
        """
        :param logic: the logical operator for the query, one of :class:`LogicalOperator`.
        """
        self.logic = logic
        self.terms = []
        self.facets = []
        self.queries = []

    def term(self, term: str):
        """
        Add a term to the query.

        :param term: the term to add to the query.
        """
        self.terms.append(term)

    def facet(self, field: str, comp: ComparisonOperator, value):
        """
        Add a facet filter to the query.

        :param field: the name of the field to filter by.
        :param comp: the comparison operator to use, one of :class:`ComparisonOperator`.
        :param value: the value to filter by.
        """
        self.facets.append((field, comp, value))

    def query(self, query: 'Query'):
        """
        Add a nested query to the query.

        :param query: the query object to add.
        """
        self.queries.append(query)

    def to_dict(self) -> typing.Dict[str, object]:
        """
        Transform the query object into a dictionary.

        :return: a dictionary object for the query.
        """
        result = {'logic': self.logic.value}

        if len(self.terms) > 0:
            result['terms'] = self.terms

        if len(self.facets) > 0:
            result['facets'] = [[x[0], x[1].value, x[2]] for x in self.facets]

        if len(self.queries) > 0:
            result['queries'] = [x.to_dict() for x in self.queries]

        return result

    def to_str(self) -> str:
        """
        Transform the query object into a string.

        :return: a string for the query.
        """
        result = []

        for x in self.terms:
            result.append(repr(x))

        for x in self.facets:
            result.append('{0}{1}{2}'.format(x[0], x[1].value, repr(x[2])))

        for x in self.queries:
            result.append('({0})'.format(x.to_str()))

        return ' {0} '.format(self.logic.value).join(result)

    def __str__(self) -> str:
        """
        Implementation of the :func:`str` function.
        """
        return self.to_str()

    def __eq__(self, other: object) -> bool:
        """
        Implementation of the operator for comparison.
        """
        return str(self) == str(other)


class Search(object):
    """
    The search class is used to represent a search request. A search request is
    usually comprised of a query and other options, such a sorting, slicing, etc.
    """
    def __init__(self):
        """
        :var list queries: a list of subqueries to this one, or `None`.
        :var list sorts: a list of sorting tuples for the query, in order
            of priority.
        :var tuple slices: the slice (offset and limit) to apply on the result set.
        :var list facets: a list of facets to return. `None` means all defined facets
            are returned. An empty list means no facets are returned.
        :var list doc_types: a list of document types to search against, or `None` for
            all document types.
        """
        self.queries = None
        self.sorts = []
        self.slices = (0, 10)
        self.facets = None
        self.doc_types = None

    def query(self, query: Query):
        """
        Add a query to the search object.

        :param query: the query object to add.
        """
        self.queries = query

    def sort(self, field: str, order: SortOrder=SortOrder.ASC):
        """
        Add a sort to the search object.

        :param field: the field to sort by.
        :param order: the sort order to use for that field, one of :class:`SortOrder`.
        """
        self.sorts.append((field, order))

    def slice(self, offset: int, limit: int):
        """
        Add a slice to the search object.

        :param offset: the offset of the slice.
        :param limit: the maximum number of results to return.
        """
        self.slices = (offset, limit)

    def facet(self, *names):
        """
        Add or disable facets in the search object.

        :param *names: a list of facets to return. When empty, disables facets completely.
        """
        self.facets = names

    def type(self, *names):
        """
        Set the document types for the search object.

        :param *names: a list of document types to search into.
        """
        self.doc_types = names

    def to_dict(self) -> typing.Dict[str, object]:
        """
        Transform the search object into a dictionary.

        :return: the search object as a dictionary.
        """
        result = {
            'offset': self.slices[0],
            'limit': self.slices[1],
            'type': self.doc_types,
            'facet': self.facets,
        }

        if self.queries is not None:
            result['query'] = self.queries.to_dict()

        if len(self.sorts) > 0:
            result['sort'] = self.sorts

        return result

    def to_str(self) -> str:
        """
        Tranform the search object into a string.

        :return: the search string.
        """
        result = []

        if self.queries is not None:
            result.append(self.queries.to_str())

        if len(self.sorts) > 0:
            for x in self.sorts:
                result.append('with sort({0}, {1})'.format(*x))

        if self.facets is not None:
            result.append('with facet({0})'.format(', '.join(self.facets)))

        if self.doc_types is not None:
            result.append('with type({0})'.format(', '.join(self.doc_types)))

        result.append('with slice({0}, {1})'.format(*self.slices))

        return ' '.join(result)

    def __str__(self) -> str:
        """
        Implementation of the :func:`str` function.
        """
        return self.to_str()

    def __eq__(self, other: object) -> bool:
        """
        Implementation of the comparison operator.
        """
        return str(self) == str(other)


class QueryParser(object):
    """
    This class provides an interface for parsing queries and search requests.
    """
    def loads(self, data: str, *, query_only: bool=False) -> (Search, Query):
        """
        Parse a query or a search request.

        :param data: the input request data.
        :param query_only: when `True`, only the query will be parsed. `False` will
            parse additional search options as well.
        :return: a search object if `query_only` is `False`, otherwise a query object.
        :raises: :class:`QueryParserError`: when a parsing error occurs.
        """
        logger.info('received query', query_only=query_only, data=data)

        tokens = tokenize.tokenize(io.BytesIO(data.encode('utf-8')).readline)
        tokens = list(tokens)

        search = Search()
        query = Query(None)

        buffers = []
        nesting = 0
        negated = False

        comp_operators = ['==', '>', '<', '>=', '<=', '!=']

        for i, token in enumerate(tokens[1:]):
            try:
                next_token = tokens[i+2]
            except IndexError:
                next_token = None

            logger.debug('token',
                         id=id(token), type=token.type, string=token.string,
                         start_x=token.start[0], start_y=token.start[1],
                         end_x=token.end[0], end_y=token.end[1])

            # Option start
            if token.type == tokenize.NAME and token.string == 'with':
                logger.debug('option start', id=id(token))

                if nesting > 0 or len(buffers) > 0:
                    raise QueryParserError(token)

                buffers.append(token.string)

            elif 'with' in buffers:
                # Function arguments
                if token.type == tokenize.OP and token.string in ['(', ',', '.']:
                    logger.debug('function argument', id=id(token))
                    buffers.append(token.string)

                # Function end
                elif token.type == tokenize.OP and token.string == ')':
                    logger.debug('function end', id=id(token))

                    buffers.append(token.string)
                    func = buffers[1]
                    bufs = []
                    args = []

                    for x in buffers[3:]:
                        if x in [',', ')']:
                            try:
                                if len(bufs) > 1:
                                    val = ''.join(bufs)
                                else:
                                    val = bufs[0]
                                args.append(val)
                            except IndexError:
                                pass
                            bufs.clear()
                        elif x == '.':
                            bufs.append(x)
                        else:
                            try:
                                val = eval(x)
                                bufs.append(val)
                            except NameError:
                                bufs.append(x)

                    try:
                        getattr(search, func)(*args)
                    except TypeError as e:
                        raise QueryParserError(str(e))
                    except AttributeError:
                        raise QueryParserError('unknown option {0}'.format(func))

                    buffers.clear()

                # Function name
                elif token.type == tokenize.NAME and len(buffers) == 1:
                    logger.debug('function name', id=id(token))
                    buffers.append(token.string)

                elif token.type in [tokenize.NAME, tokenize.NUMBER]:
                    logger.debug('function argument', id=id(token))
                    buffers.append(token.string)

                else:
                    raise QueryParserError(token)

            elif (token.type == tokenize.NAME and nesting == 0 and
                  token.string not in ('None', 'True', 'False')):
                # Logical operator
                if token.string == 'and':
                    logger.debug('logical operator', id=id(token))
                    if query.logic is not None and query.logic != LogicalOperator.AND:
                        raise QueryParserError(token)
                    query.logic = LogicalOperator(token.string)

                # Logical operator
                elif token.string == 'or':
                    logger.debug('logical operator', id=id(token))
                    if query.logic is not None and query.logic != LogicalOperator.OR:
                        raise QueryParserError(token)
                    query.logic = LogicalOperator(token.string)

                # Logical operator
                elif token.string == 'not':
                    logger.debug('logical operator', id=id(token))
                    if query.logic is not None and query.logic != LogicalOperator.NOT:
                        raise QueryParserError(token)
                    query.logic = LogicalOperator(token.string)

                # Other operator or field name
                else:
                    logger.debug('field name', id=id(token))
                    buffers.append(token.string)

            elif token.type == tokenize.OP:
                # Sub queries start
                if token.string == '(':
                    logger.debug('sub query start', id=id(token))
                    nesting += 1

                    if nesting > 1:
                        buffers.append(token.string)

                # Sub queries end
                elif token.string == ')':
                    logger.debug('sub query end', id=id(token))
                    nesting -= 1

                    if nesting > 0:
                        buffers.append(token.string)
                    else:
                        value = ' '.join(buffers)
                        sub = self.loads(value, query_only=True)
                        query.query(sub)
                        buffers.clear()

                # Facet nested names
                elif token.string == '.' and nesting == 0:
                    logger.debug('nested facet', id=id(token))
                    buffers.append(token.string)

                # Comparison operators
                elif token.string in comp_operators and nesting == 0:
                    logger.debug('comparison operator', id=id(token))
                    buffers.append(token.string)

                elif (token.string == '-' and nesting == 0 and
                      next_token.type == tokenize.NUMBER):
                    logger.debug('negative number operator', id=id(token))
                    negated = True

                elif nesting == 0:
                    raise QueryParserError(token)

                else:
                    logger.debug('value', id=id(token))
                    buffers.append(token.string)

            elif (token.type in [tokenize.NAME, tokenize.STRING, tokenize.NUMBER] and
                  nesting == 0):
                # Terms
                if len(buffers) == 0:
                    logger.debug('value', id=id(token))
                    value = eval(token.string)
                    query.term(value)
                    buffers.clear()

                # Facet values
                elif len(set(buffers).intersection(comp_operators)) > 0:
                    logger.debug('as facet value', id=id(token))
                    field = ''.join(buffers[0:-1])
                    comp = buffers[-1]
                    value = eval(token.string)
                    if negated:
                        value = 0 - value
                        negated = False
                    query.facet(field, ComparisonOperator(comp), value)
                    buffers.clear()

                else:
                    raise QueryParserError(token)

            elif nesting > 0:
                buffers.append(token.string)

            elif token.type != 0 or len(buffers) > 0 or nesting > 0:
                raise QueryParserError(token)

        if query.logic is None:
            query.logic = LogicalOperator.AND

        if query_only:
            return query

        search.query(query)
        return search


class QueryParserError(SyntaxError):
    """
    This exception class is raised when the :class:`QueryParser` encounters an error
    during parsing.
    """
    def __init__(self, token: typing.Tuple[str, tokenize.TokenInfo]):
        """
        :param token: a message or the token object that caused the error.
        """
        if not isinstance(token, str):
            super().__init__('unexpected token "{0}" at offset {1}'
                             .format(token.string, token.start[1]))
        else:
            super().__init__(token)



"""
This module provides generic manager objects, that is, objects that can create or save objects
of a specific type and resolve them later.
"""

import typing


class ResourceManager(object):
    """
    The manager class can create, store and resolve objects of a specific type
    later, much like a dictionary, but with additional functionality such as
    duplicates handling.
    """
    def __init__(self, cls: type):
        """
        :param cls: the type of object that the manager manages.
        """
        self.cls = cls
        self.stack = {}

    def new(self, _name: str, *args, **kwargs) -> object:
        """
        Create a new object in the manager.

        This method basically instantiate the object type defined when the manager
        is created using the arguments and keyword arguments specified. It then
        saves it into the manager stack and can be resolved later by its name.

        :param _name: the name to give the instance in the manager, must be unique.
        :param *args: the arguments passed to the managed class initializer.
        :param **kwargs: the keyword arguments passed to the managed class initializer.
        :return: an instance of the managed class.
        :raises: :class:`ResourceExistsError`: when the specified `name` already exists
            in the manager.
        """
        if _name in self.stack:
            raise ResourceExistsError(_name)

        result = self.cls(*args, **kwargs)
        self.stack[_name] = result
        return result

    def save(self, name: str, inst: object) -> object:
        """
        Save an instance in the manager.

        :param name: the name to give the instance in the manager, must be unique.
        :param inst: the instance of the object to save.
        :return: the managed instance object.
        :raises: :class:`ResourceExistsError`: when the specified `name` already exists
            in the manager.
        """
        if name in self.stack:
            raise ResourceExistsError(name)

        self.stack[name] = inst
        return inst

    def save_all(self, insts: typing.Dict[str, object]):
        """
        Save multiple instances at once in the manager.

        This method is basically a the :meth:`save` method wrapped in a loop.

        :param insts: a dictionary of instances to save, with the key being the name
            to give the instance, and the value being the instance itself.
        :raises: :class:`ResourceExistsError`: when one of the specified entries
            has a name that already exists. This method is safe, so if the exception
            is raised, it can be assumed that none of the objects have been saved and
            the manager state has not been modified.
        """
        for name in insts:
            if name in self.stack:
                raise ResourceExistsError(name)

        for name, inst in insts.items():
            self.save(name, inst)

    def resolve(self, name: str) -> object:
        """
        Resolve an existing object instance.

        :param name: the name of the object to resolve.
        :return: the managed object with that name.
        :raises: :class:`ResourceNotFoundError`: if the object has not been registered.
        """
        try:
            return self.stack[name]
        except KeyError:
            raise ResourceNotFoundError(name)


class ResourceExistsError(NameError):
    """
    This exception is raised when trying to save or create an object instance in a
    manager, when another object with the same name already exists in that manager.
    """
    pass


class ResourceNotFoundError(NameError):
    """
    This exception is raised when trying to resolve an object instance with a name
    that has not been registered in the manager.
    """
    pass



"""
This module provides some utilities and classes to implement multiprocessing workloads.
"""

import abc
import signal
import typing
import threading
import multiprocessing
from queue import Queue
from queue import Empty

from relevance import loggers


# Logging
logger = loggers.getLogger('relevance.worker')


class Worker(object, metaclass=abc.ABCMeta):
    """
    This interface allows to implement different worker types, such as threaded workers,
    process workers or network workers.
    """
    def __init__(self):
        """
        :var scope: contains the signal queue for the worker.
        :var children: contains the list of children workers for this worker.
        :var is_started: whether the worker is started or not.
        """
        self.scope = Queue()
        self.handles = {}
        self.parents = []
        self.children = []
        self.is_started = False
        logger.debug('init', inst=self)

    @abc.abstractmethod
    def run(self):
        """
        This is the method that needs to be implemented based on the worker type.
        For threaded workers per example, this method will spawn a thread and start it.
        """
        pass

    def add(self, worker: 'Worker'):
        """
        Add a child worker.

        Child workers can emit signals down to their parents or receive signal broadcasts
        from their parents.

        :param worker: the child worker object.
        :raises: :class:`InvalidWorkerStateError`: if the worker is started. A worker's
            hierarchy cannot be modified once it's started. It needs to be stopped first,
            then it can be modified.
        """
        if self.is_started:
            raise InvalidWorkerStateError()

        worker.parents.append(self)
        self.children.append(worker)
        logger.debug('add', inst=self, worker=worker)

    def signal(self, signal: str, **args):
        """
        Send a signal to the worker.

        This only sends a signal to the current worker. To send signals up or down the
        worker's hierarchy, see :meth:`Worker.emit`, :meth:`Worker.broadcast` and
        :meth:`Worker.trigger`.

        :param signal: the signal to emit.
        :param args: the signal arguments.
        """
        if self.scope is None:
            return

        logger.info('signal', inst=self, signal=signal, args=args)
        self.scope.put((signal, args))

    def emit(self, signal: str, **args):
        """
        Emit a signal to the worker's parents.

        This method does not send a signal to the current worker however, only to
        its parents.

        :param signal: the signal to emit.
        :param args: the signal arguments.
        """
        logger.debug('emit', inst=self, signal=signal, args=args)
        for x in self.parents:
            x.signal(signal, **args)
            x.emit(signal, **args)

    def broadcast(self, signal: str, **args):
        """
        Broadcast a signal up to all the worker's children.

        This method does not send a signal to the current worker, only its children.
        To send a signal to the current worker and its children as well, use
        :meth:`Worker.trigger`.

        :param signal: the signal to emit.
        :param args: the signal arguments.
        """
        logger.debug('broadcast', inst=self, signal=signal, args=args)
        for x in self.children:
            x.signal(signal, **args)
            x.broadcast(signal, **args)

    def trigger(self, signal: str, **args):
        """
        Send a signal to the current worker and all of its children.

        This method sends a signal to the current worker's children and itself.
        To only send a signal to the children, use :meth:`Worker.broadcast`.

        :param signal: the signal to emit.
        :param args: the signal arguments.
        """
        logger.debug('trigger', inst=self, signal=signal, args=args)
        self.signal(signal, **args)
        self.broadcast(signal, **args)

    def listen(self, signal: str, func: typing.Callable, **args):
        """
        Add a signal handler in the worker.

        :param signal: the signal to handle.
        :param func: a callable to execute when the signal is received.
        :param args: default arguments to pass to the callable. Whenever the specified
            signal is received, the arguments specified here will be passed to all
            handlers whether they were in the payload or not.
        :raises: :class:`InvalidWorkerStateError`: if the worker is started. A worker's
            handlers cannot be modified once it's started. It needs to be stopped first,
            then it can be modified.
        """
        if self.is_started:
            raise InvalidWorkerStateError()

        logger.debug('listen', inst=self, signal=signal, func=func, args=args)
        if signal not in self.handles:
            self.handles[signal] = []
        self.handles[signal].append((func, args))

    def execute(self, signal: str, **args):
        """
        Execute all registered handlers for the specified signal in the current worker.

        This method is mostly use internally. To generate a signal, use :meth:`Worker.signal`.

        :param signal: the signal to trigger.
        :param args: the signal arguments.
        """
        if signal in self.handles:
            for func, kwargs in self.handles[signal]:
                args = {**args, **kwargs}
                func(**args)

    def event_loop(self):
        """
        The worker event loop.

        This method is mostly used internally. You should not override it unless you know
        what you're doing. It takes care of receiving and handling the multiple signals
        that are sent to the worker, and cleaning up after it stops.
        """
        self.execute('worker_start')

        while self.is_started or not self.scope.empty():
            self.execute('loop_start')

            try:
                signal, args = self.scope.get(timeout=0.001)
                self.scope.task_done()
                self.execute(signal, **args)

                if signal == 'stop':
                    if args['force']:
                        break
                    else:
                        self.is_started = False
            except Empty:
                pass

            self.execute('loop_end')

        self.is_started = False
        self.execute('worker_end')

    def start(self):
        """
        Start the worker.

        If the worker is already started, this has no effect.
        """
        for x in self.children:
            x.start()

        if self.is_started:
            return

        logger.debug('start', inst=self)
        self.is_started = True
        self.run()

    def stop(self, force: bool=False):
        """
        Stop the worker.

        When requesting a worker to stop, it will stop only when its event queue is
        empty. Using `force` will avoid that and the worker will be stopped as soon as
        the current event, if any, finishes.

        If the worker is already stopped, this has no effect.

        :param force: whether to force stop the worker or not.
        """
        logger.debug('stop', inst=self, force=force)
        self.trigger('stop', force=force)

    def kill(self):
        """
        Kill the worker.

        This should be rarely used, but can be useful to terminating stuck workers and
        the likes. Unlike :meth:`Worker.stop`, this method does not perform any clean up.
        Caution is advised.
        """
        logger.debug('kill', inst=self)
        for x in self.children:
            x.kill()

        self.is_started = False
        while not self.scope.empty():
            self.scope.get()
            self.scope.task_done()

    def join(self):
        """
        Join the worker.

        Calling this method will wait the current worker has stopped, as well as
        all its children. Make sure to call :meth:`Worker.stop` first.
        """
        logger.debug('join', inst=self)
        for x in self.children:
            x.join()


class WorkerProxy(object):
    """
    This class is useful for implementing workers within other objects. It proxies
    the :class:`Worker` interface through composition.
    """
    def __init__(self, worker: Worker):
        """
        :param worker: the worker object to use for the instance.
        """
        self.worker = worker

    def emit(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.emit`.
        """
        self.worker.emit(*args, **kwargs)

    def broadcast(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.broadcast`.
        """
        self.worker.broadcast(*args, **kwargs)

    def trigger(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.trigger`.
        """
        self.worker.trigger(*args, **kwargs)

    def listen(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.listen`.
        """
        self.worker.listen(*args, **kwargs)

    def start(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.start`.
        """
        self.worker.start(*args, **kwargs)

    def stop(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.stop`.
        """
        self.worker.stop(*args, **kwargs)

    def kill(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.kill`.
        """
        self.worker.kill(*args, **kwargs)

    def join(self, *args, **kwargs):
        """
        Implementation of :meth:`Worker.join`.
        """
        self.worker.join(*args, **kwargs)


class SimpleWorker(Worker):
    """
    This simple worker implementation runs in the current thread and does not
    multiprocess at all. Starting it will block the current process.

    However, when the current process receives the `SIGINT` signal, the worker
    will generate a `stop` signal and be allowed to stop.
    """
    def run(self):
        """
        Implementation of :meth:`Worker.run`.
        """
        signal.signal(signal.SIGINT, lambda *a: setattr(self, 'is_started', False))
        self.event_loop()

    def join(self):
        """
        Implementation of :meth:`Worker.join`.
        """
        super().join()
        while self.is_started or not self.scope.empty():
            self.event_loop()
        self.scope = None


class ThreadWorker(Worker):
    """
    This worker class uses a thread for multiprocessing.
    """
    def __init__(self):
        """
        .. init
        """
        super().__init__()
        self.thread = None

    def run(self):
        """
        Implementation of :meth:`Worker.run`.
        """
        if self.thread is not None:
            raise InvalidWorkerStateError()

        self.thread = threading.Thread(target=self.event_loop)
        self.thread.start()

    def kill(self):
        """
        Implementation of :meth:`Worker.kill`.
        """
        if self.thread is None:
            raise InvalidWorkerStateError()

        logger.debug('kill', inst=self)
        self.thread.terminate()

    def join(self):
        """
        Implementation of :meth:`Worker.join`.
        """
        if self.thread is None:
            raise InvalidWorkerStateError()

        super().join()
        self.thread.join()
        self.thread = None


class ProcessWorker(Worker):
    """
    This worker type uses spawned processes for multiprocessing. Spawned process
    do not shared any resource with their parents: they are isolated.
    """
    context_type = 'spawn'
    """ :const: The multiprocessing context type for the worker. """

    def __init__(self):
        """
        .. init
        """
        super().__init__()
        self.process = None
        self.context = multiprocessing.get_context(self.context_type)
        self.scope = self.context.JoinableQueue()

    def run(self):
        """
        Implementation of :meth:`Worker.run`.
        """
        if self.process is not None:
            raise InvalidWorkerStateError()

        self.process = self.context.Process(target=self.event_loop)
        self.process.start()

    def kill(self):
        """
        Implementation of :meth:`Worker.kill`.
        """
        if self.process is None:
            raise InvalidWorkerStateError()

        logger.debug('kill', inst=self)
        for x in self.children:
            x.kill()

        self.process.terminate()

    def join(self):
        """
        Implementation of :meth:`Worker.join`.
        """
        if self.process is None:
            raise InvalidWorkerStateError()

        super().join()
        self.process.join()
        self.process = None
        self.scope = None


class ForkedWorker(ProcessWorker):
    """
    Like a :class:`ProcessWorker`, but forked workers can actually share resources
    with their parents, such as file descriptors.
    """
    context_type = 'fork'
    """ :const: The multiprocessing context type for the worker. """


class InvalidWorkerStateError(RuntimeError):
    """
    This exception is raised when attempting to perform an operation, with the worker
    not being in a valid state for that specific operation. Per example, attempting
    to add a new signal handler to a started worker will cause this error.
    """
    pass



"""
This service module provides the API for searchers and engines.
"""

from relevance.service import Service
from relevance.settings import SettingsFactory
from relevance.manager import ResourceManager
from relevance.query import QueryParser
from relevance.engine import Engine
from relevance.engine import EngineFactory


# Initialize the service
service = Service.instance('search')
manager = service.data.manager = ResourceManager(Engine)


@service.before_first_request
def _bootstrap():
    """
    Bootstrap the service.
    """
    factory = EngineFactory()
    loader = SettingsFactory(['/etc/relevance', './etc'])
    settings = service.data.settings = loader.load('search')

    engines = settings.get('engines', {}, data_type=dict)
    engines = factory.load_all_engines(engines)

    for x in engines.values():
        x.start()

    manager.save_all(engines)


@service.route('/<engine_name>', methods=['GET'])
def do_search(engine_name: str):
    """
    Perform a search request.

    :Method: GET
    :Path: /engines/`engine_name`
    :Query Arguments: **q** - the search request to execute. See :mod:`relevance.query`.
    :param engine_name: the name of the engine to query.

    :Statuses: `200` on success.
    :Response:
        .. code-block:: json

            {
                "search": {
                    "query": "foo==\"bar\"",
                },
                "results": [
                    {"text": "my test document"},
                    ...
                ],
                "facets": {
                    "foo": {
                        "bar": 42,
                        "baz": 12
                    }
                },
                "time": 0.0432,
                "count": 143
            }
    """
    engine = manager.resolve(engine_name)

    query = service.request.args.get('q', '')
    parser = QueryParser()
    search = parser.loads(query)

    result_set = engine.search(search)
    result = {
        'search': search.to_dict(),
        'results': result_set,
        'time': result_set.time,
        'count': result_set.count,
    }

    if result_set.facets is not None:
        result.update({
            'facets': result_set.facets,
        })

    return service.result(200, result)


if __name__ == '__main__':
    import os
    service.run(port=int(os.getenv('PORT', 55345)), debug=True)



"""
This module provides various utilities for command line tools.
"""

import sys
import pydoc
import logging
import argparse
# from urllib.parse import urlencode
from twisted.web.resource import Resource
from twisted.web.resource import NoResource
# from twisted.web.wsgi import WSGIResource
# from twisted.web.server import Site
# from twisted.internet import reactor
from twisted.python.log import PythonLoggingObserver

import relevance


# Loggers
logging.basicConfig(stream=sys.stderr)
observer = PythonLoggingObserver()
observer.start()
logger = logging.getLogger('relevance')
logger.setLevel(logging.INFO)
logging.getLogger().setLevel(logging.ERROR)


class _RootResource(Resource):
    """
    Twisted root resource implementation.

    This class allows to bind a single twisted reactor to multiple WSGI applications.
    """
    def getChild(self, child, request):
        """
        Get a child application.
        """
        name = request.prepath.pop().decode('ascii')
        return self.children.get(name, NoResource())


def _args_default(parser: argparse.ArgumentParser):
    """
    Add some default arguments to the parser.

    :param parser: the parser object.
    """
    parser.add_argument(
        '--version',
        help='show the version information',
        action='store_true',
    )
    parser.add_argument(
        '-q', '--quiet',
        help='only log warnings and errors',
        action='store_const',
        dest='verbosity',
        const=0,
    )
    parser.add_argument(
        '-qq',
        help='only log errors',
        action='store_const',
        dest='verbosity',
        const=-1,
    )
    parser.add_argument(
        '-v', '--verbose',
        help='enable verbose logging',
        action='store_const',
        dest='verbosity',
        const=1,
    )
    parser.add_argument(
        '-vv',
        help='enable extra verbose logging',
        action='store_const',
        dest='verbosity',
        const=2,
    )
    parser.add_argument(
        '-vvv',
        help='enable ultra verbose logging',
        action='store_const',
        dest='verbosity',
        const=3,
    )


def _args_handle(args: argparse.Namespace):
    """
    Handle some default arguments.

    :param args: the argument namepsace object.
    """
    if args.version:
        print('Relevance {0}'.format(relevance.__version__))
        sys.exit(0)

    # Configure the loggers
    if args.verbosity == -1:
        logger.setLevel(logging.ERROR)

    if args.verbosity == 0:
        logger.setLevel(logging.WARNING)

    if args.verbosity == 1:
        logger.setLevel(logging.DEBUG)
        logging.getLogger().setLevel(logging.WARNING)

    if args.verbosity == 2:
        logger.setLevel(logging.DEBUG)
        logging.getLogger().setLevel(logging.INFO)

    if args.verbosity == 3:
        logger.setLevel(logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)


# def server():
#     """
#     Start one or more web application server.
#     """
#     parser = argparse.ArgumentParser(
#         prog='relevance-server',
#         description='Start Relevance web application server.',
#     )
#     parser.add_argument(
#         '-a',
#         help='set the applications to start',
#         metavar='MODULE:PATH',
#         nargs='+',
#         dest='apps',
#         default=['search'],
#     )
#     parser.add_argument(
#         '-s',
#         help='set the listener to use for the server',
#         metavar='ADDR:PORT',
#         default='0.0.0.0:5346',
#         dest='listen',
#     )
#     _args_default(parser)
#     args = parser.parse_args(sys.argv[1:])
#     _args_handle(args)
#
#     try:
#         logger.info('version {0}'.format(relevance.__version__))
#
#         # Add the selected applications
#         root = _RootResource()
#
#         for spec in args.apps:
#             parts = spec.split(':')
#
#             if len(parts) == 1:
#                 parts.append(parts[0])
#
#             if len(parts) != 2:
#                 raise argparse.ArgumentTypeError('invalid server spec')
#
#             module, path = parts
#             app_name = 'relevance.api.{0}.app'.format(module)
#
#             app = pydoc.locate(app_name)
#             if app is None:
#                 raise argparse.ArgumentTypeError('invalid application {0}'.format(module))
#
#             start_app(app)
#
#             if len(args.apps) > 1:
#                 resource = WSGIResource(reactor, reactor.getThreadPool(), app)
#                 root.putChild(path, resource)
#                 logger.info('binding module {0} to /{1}/'.format(
#                     module, path,
#                 ))
#             else:
#                 root = WSGIResource(reactor, reactor.getThreadPool(), app)
#                 logger.info('binding module {0} to /'.format(
#                     module,
#                 ))
#
#         # Setup the reactor
#         try:
#             addr, port = args.listen.split(':')
#             port = int(port)
#         except (ValueError, TypeError):
#             raise argparse.ArgumentTypeError('invalid listener spec')
#
#         logger.info('starting application server on {0}:{1}'.format(
#             addr, port,
#         ))
#         reactor.listenTCP(port, Site(root), interface=addr)
#         reactor.run()
#
#         print()
#         logger.info('server shutting down')
#
#     except argparse.ArgumentTypeError as e:
#         logger.critical('startup error: {0}'.format(
#             str(e),
#         ))
#         sys.exit(1)
#
#     except Exception as e:
#         logger.critical('{0}: {1}'.format(e.__class__.__name__, str(e)))
#         sys.exit(1)
#
#
# def search():
#     """
#     Perform a search.
#     """
#     from relevance.api import search as api
#
#     parser = argparse.ArgumentParser(
#         prog='relevance-search',
#         description='Perform a search request locally on a Relevance engine',
#     )
#     parser.add_argument(
#         'engine_name',
#         help='the name of the engine to query',
#         metavar='ENGINE_NAME',
#         nargs='?',
#     )
#     parser.add_argument(
#         'query',
#         help='the search request to execute',
#         metavar='QUERY',
#         nargs='?',
#         default='',
#     )
#     parser.add_argument(
#         '--list',
#         help='list the configured engines',
#         action='store_true',
#     )
#     _args_default(parser)
#     args = parser.parse_args()
#     _args_handle(args)
#
#     start_app(api.app)
#
#     try:
#         logger.info('version {0}'.format(relevance.__version__))
#
#         environ = {
#             'SERVER_NAME': 'localhost',
#             'SERVER_PORT': '0',
#             'REQUEST_METHOD': 'GET',
#             'QUERY_STRING': urlencode({'q': args.query}),
#             'wsgi.url_scheme': 'http',
#         }
#
#         with api.app.request_context(environ):
#             if args.list:
#                 response = api.get_config()
#                 print(response.data.decode('utf-8'))
#             else:
#                 if not args.engine_name:
#                     raise argparse.ArgumentTypeError('need to specify an engine name')
#
#                 response = api.do_search(args.engine_name)
#                 print(response.data.decode('utf-8'))
#
#     except argparse.ArgumentTypeError as e:
#         logger.critical('utility error: {0}'.format(
#             str(e),
#         ))
#         sys.exit(1)
#
#     except Exception as e:
#         logger.critical('{0}: {1}'.format(e.__class__.__name__, str(e)))
#         sys.exit(1)


def _main():
    """
    Master command line tool.
    """
    if '--version' in sys.argv:
        print('Relevance {0}'.format(relevance.__version__))
        sys.exit(0)

    try:
        prog = sys.argv.pop(1)
        if prog[0] == '_':
            raise TypeError()

        name = '{0}.{1}'.format(_main.__module__, prog)
        cmd = pydoc.locate(name)
        if cmd is None:
            raise TypeError()
    except (TypeError, IndexError):
        print('usage: relevance --version')
        # print('       relevance server [<options>]')
        # print('       relevance search [<options>]')
        sys.exit(1)
    cmd()


if __name__ == '__main__':
    _main()



"""
This module provides the interfaces for loading settings from configuration files
into useable objects.
"""

import os
import typing
import anyconfig


class Settings(dict):
    """
    This class represents a settings configuration.

    Although it can be instantiated manually, it is typically created using a
    :class:`SettingsFactory` instance.
    """
    def __init__(self, name: str, data: typing.Iterable=''):
        """
        :param name: the name to give this settings subset.
        :param data: the iterable to import as data.
        """
        super().__init__(data)
        self.name = name

    def __getitem__(self, key: str) -> object:
        """
        Get a settings item in dictionary fashion.

        This method is mostly used if no validation of the data type is required.
        If validation is required, use the :meth:`Settings.get_type` method instead.

        :returns: the found setting object. If the result is a :class:`dict`, it will
        be automatically converted into a :class:`Settings` object.
        :raises: :class:`RequiredSettingError`: if the setting is missing.
        """
        name = '{0}/{1}'.format(self.name, key)

        try:
            item = super().__getitem__(key)
        except KeyError as e:
            raise RequiredSettingError(str(e))

        if not isinstance(item, dict):
            return item

        return Settings(name, item)

    def get_type(self, key: str, data_type: type) -> object:
        """
        Get a setting and validate its type.

        :param key: the setting key to retrieve.
        :param data_type: a data type to validate against, such as :class:`str`.
        :returns: the found setting object. If the result is a :class:`dict`, it will
            be automatically converted into a :class:`Settings` object.
        :raises: :class:`InvalidSettingTypeError`: if the data type is invalid.
        """
        name = '{0}/{1}'.format(self.name, key)
        item = self[key]

        if not isinstance(item, data_type):
            raise InvalidSettingTypeError(key, data_type)

        if not isinstance(item, dict):
            return item

        return Settings(name, item)

    def get(self, key: str, default: object=None, data_type: type=None) -> object:
        """
        Get an optional settings and validate its type.

        If the requested settings is not found, the `default` value is returned.

        :param data_type: an optional data type to validate against, e.g.: :class:`str`.
        :returns: the found setting object. If the result is a :class:`dict`, it will
            be automatically converted into a :class:`Settings` object.
        :raises: :class:`InvalidSettingTypeError`: if the data type is invalid.
        """
        name = '{0}/{1}'.format(self.name, key)
        item = super().get(key, default)

        if data_type is not None and not isinstance(item, data_type) and default is not None:
            raise InvalidSettingTypeError(key, data_type)

        if not isinstance(item, dict):
            return item

        return Settings(name, item)

    def resolve(self, path: str) -> object:
        """
        Resolve a nested dictionary key path.

        When manipulating settings objects with multiple levels, this method allows
        to resolve a sub-object quickly, using slashes as separators.

        :param path: the path to resolve.
        :returns: the resolved value.
        """
        parts = path.split('/')
        ref = self
        for x in parts:
            ref = ref[x]
        return ref

    def first_item(self, data_type: type=None) -> tuple:
        """
        Get the first item.

        :param data_type: an optional data type to validate against.
        :returns: a tuple containing the key and the value of the first item
            in the settings object.
        :raises: InvalidSettingTypeError: if the data type is invalid.
        """
        try:
            key, value = list(self.keys())[0], list(self.values())[0]
        except TypeError:
            raise RequiredSettingError()

        if data_type is not None and not isinstance(value, data_type):
            raise InvalidSettingTypeError(key, data_type)

        if not isinstance(value, dict):
            return (key, value)

        return (key, Settings(key, value))

    def items(self) -> typing.Iterable:
        """
        Implementation of :meth:`dict.__iter__`.
        """
        for x, y in super().items():
            if not isinstance(y, dict):
                yield (x, y)
            else:
                yield (x, Settings(x, y))


class SettingsFactory(object):
    """
    This class allows to load one or multiple files, based on environment, in
    specific directories as settings objects.
    """
    def __init__(self, dirs: list, default_env: str='devel', env: str=None):
        """
        :param dirs: a list of directories to lookup, in reverse priority order
            (the last one overrides the first one).
        :param default_env: the default environment to use if none is set. The
            factory uses the `PYTHON_ENV` environment variable to determine the
            name of the environment.
        :param env: the environment to enforce, whether it is set or not.
        """
        self.dirs = dirs
        self.env = env if env is not None else os.getenv('PYTHON_ENV', default_env)

    def load(self, pattern: str) -> Settings:
        """
        Load a settings object from configuration files.

        :param pattern: the string pattern to match, i.e.: the base name of the
            configuration filename. Per example, given a `pattern` of `foo`, and
            an environment name of `bar`, the following files will be attempted in
            this order: `foo.json`, `foo.yml`, `foo.bar.json`, `foo.bar.yml`.
        :returns: a settings object.
        """
        filenames = []

        for dirname in self.dirs:
            filenames += [
                '{0}/{1}.json'.format(dirname, pattern),
                '{0}/{1}.yml'.format(dirname, pattern),
                '{0}/{1}.{2}.json'.format(dirname, pattern, self.env),
                '{0}/{1}.{2}.yml'.format(dirname, pattern, self.env),
            ]

        data = anyconfig.load(filenames, ignore_missing=True)
        return Settings(pattern, data)


class RequiredSettingError(KeyError):
    """
    This exception is raised when attempting to retrieve a settings by its key
    and the key does not exist in the settings object.
    """
    pass


class InvalidSettingTypeError(TypeError):
    """
    This exception is raised when retrieving a setting, but the returned value
    does not have the expected data type.
    """
    def __init__(self, key: str, data_type: type):
        """
        :param key: the name of the setting that was retrieved.
        :param data_type: the expected data type.
        """
        super().__init__('setting {0} expected to be a {1}'.format(
            key, data_type,
        ))
        self.key = key
        self.data_type = data_type


class InvalidSettingValueError(ValueError):
    """
    This exception is raised when the value of a setting is has an unexpected
    value.
    """
    pass



"""
This module provides the interface for handling facet definitions. It contains core
facet types, such as term facets and histograms, and the abstract interface for implementing
additional facets.

New facet types should implement the :class:`Facet` interface. When doing so,
:class:`.engine.Engine` implementations will require to be extended with the new functionality
as well.
"""

import abc
import dateutil.parser
from dateutil.relativedelta import relativedelta


class Facet(object):
    """
    The class provides the interface to implement facet types. Sub-classing this class
    is not enough however, each :class:`.engine.Engine` implementation requires to be
    extended to support the new facet type as well.
    """
    def __init__(self, *, field: str=None, path: str=None):
        """
        :param field: the field for the facet.
        :param path: the path or the facet. Mutually exclusive with `field`. Used when
            a facet is nested.
        :raises: :class:`FacetDefinitionError`: if the facet is defined improperly.
        """
        if field is not None and path is not None:
            raise FacetDefinitionError('field and path are mutually exclusive')

        self.field = field
        self.path = path


class TermFacet(Facet):
    """
    This facet implementation aggregates on facet terms. This is mostly engine specific.
    """
    pass


class HistogramFacet(Facet, metaclass=abc.ABCMeta):
    """
    This interface can implemented to provide new histogram facets, which are typically
    range or interval aggregations.
    """
    @abc.abstractmethod
    def get_range_for(self, value: object) -> tuple:
        """
        Get a start and end range given a specific value.

        When given an arbitrary value, this method returns the start and end values
        matching the range the value is in.

        :param value: an arbitrary value to get the range for.
        :return: a tuple containing the start and end values for the range.
        """
        pass


class DateFacet(HistogramFacet):
    """
    This facet type aggregates on date and time by using a specific interval.
    """
    YEAR = 'year'
    """ Aggregate by year, with each value being Jan 1st. """

    QUARTER = 'quarter'
    """ Aggregate by quarter, with a value on the first day of Jan, Apr, Jul, Oct. """

    MONTH = 'month'
    """ Aggregate by month, with a value on the first day of each month. """

    WEEK = 'week'
    """ Aggregate by week, with a value each Sunday. """

    DAY = 'day'
    """ Aggregate by day, with a value at midnight on every day. """

    HOUR = 'hour'
    """ Aggregate by hour, with a value at :00:00 every hour. """

    MINUTE = 'minute'
    """ Aggregate by minute. """

    SECOND = 'second'
    """ Aggregate by second. """

    def __init__(self, *, interval: str=MONTH, **kwargs):
        """
        :param interval: one of the class constants to interval by, \
            e.g.: :const:`DateFacet.YEAR`.
        """
        super().__init__(**kwargs)
        self.interval = interval

    def get_range_for(self, value: object) -> tuple:
        """
        Implementation of :meth:`HistogramFacet.get_range_for`.

        :raises: :class:`FacetValueError`: if the value is not the start of a range.
        :raises: :class:`FacetDefinitionError`: if an invalid interval has been specified.
        """
        start = dateutil.parser.parse(value)
        prop, delta = self.interval, 1

        if self.interval == DateFacet.YEAR:
            if start.month != 1 or start.day != 1 or start.hour != 0 or \
               start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.QUARTER:
            if start.month not in [1, 4, 7, 10] or start.day != 1 or \
               start.hour != 0 or start.minute != 0 or start.second != 0:
                raise FacetValueError(value)
            prop, delta = 'month', 3

        elif self.interval == DateFacet.MONTH:
            if start.day != 1 or start.hour != 0 or \
               start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.WEEK:
            if start.weekday() != 0:
                raise FacetValueError(value)
            prop, delta = 'day', 7

        elif self.interval == DateFacet.DAY:
            if start.hour != 0 or start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.HOUR:
            if start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.MINUTE:
            if start.second != 0:
                raise FacetValueError(value)

        else:
            raise FacetDefinitionError('invalid interval {0}'.format(self.interval))

        end = start + relativedelta(**{'{0}s'.format(prop): delta})

        return start.isoformat(), end.isoformat()


class IntervalFacet(HistogramFacet):
    """
    This facet type aggregates using a specific numeric interval. Per example, if
    `100` is supplied as an interval, ranges of `0..100`, `100..200` will be
    returned. The minimum and maximum ranges depend on the actual data.
    """
    def __init__(self, *, interval: object, **kwargs):
        """
        :param interval: the numeric interval at which to group the results, typically
            an :class:`int` or a :class:`float`.
        """
        super().__init__(**kwargs)
        self.interval = interval

    def get_range_for(self, value: object) -> tuple:
        """
        Implementation of :meth:`HistogramFacet.get_range_for`.

        :raises: :class:`FacetValueError`: if the value is not the start of a range.
        """
        if (value % self.interval) > 0:
            raise FacetValueError(self, value)

        return value, value + self.interval


class RangeFacet(HistogramFacet):
    """
    The facet type aggregates using pre-defined ranges at specific values.
    """
    def __init__(self, *, ranges: dict, **kwargs):
        """
        :param ranges: a dictionary of ranges with each value being a tuple containing
            a start and an end value. Indefinite values can be specified using `None`.
        """
        super().__init__(**kwargs)
        self.ranges = ranges

    def get_range_for(self, value: object) -> tuple:
        """
        Implementation of :meth:`HistogramFacet.get_range_for`.

        :raises: :class:`FacetValueError`: if the value is not the start of a range.
        """
        try:
            start, end = self.ranges[value]
        except KeyError:
            raise FacetValueError(self, value)

        return start, end


class FacetDefinitionError(TypeError):
    """
    This exception is raised when a facet has not been defined properly. It can be
    because a required option has not been supplied or the supplied value is
    invalid, per example.
    """
    pass


class FacetValueError(ValueError):
    """
    This exception is raised specifically when a requested value is not valid for
    the facet. Per example, if configuring an engine with a :class:`IntervalFacet`
    with a specified interval of `100`, filtering with a value of `50` will raise
    this exception type.
    """
    pass


class FacetNotDefinedError(NameError):
    """
    This exception is raised when a facet is requested but has not yet been
    configured.
    """
    pass


class FacetNotSupportedWarning(UserWarning):
    """
    This warning is raised when an :class:`.engine.Engine` object has been configured
    using a facet type that it does not support.
    """
    def __init__(self, cls: type):
        """
        :param cls: the facet class that raised the warning.
        """
        super().__init__(cls.__name__)



"""
This module provides the functionality for binary and extended logging. Rather than use
the :mod:`logging` module directly and parse message strings, this module allows a message
to be logged with parameters, making it easier to parse by a computer.
"""

import logging


class Logger(logging.Logger):
    """
    This subclass of :class:`logging.Logger` rewires the logging method so that
    arbritary keyword arguments can be supplied along with an event type, so that
    log messages are easily machine readable.
    """
    def _log(self, level: int, event_type: str, *args, **payload):
        """
        Implementation of :meth:`logging.Logger._log`.

        :param level: the log message level, as defined in the :mod:`logging` module.
        :param event_type: the event type, typically a string that describes the current
            action or state to log.
        :param **payload: the keyword arguments to attach to the log message.
        """
        parts = []
        for key, value in payload.items():
            parts.append('{0}={1}'.format(key, value))

        if len(parts) > 0:
            msg = '{0} with {1}'.format(event_type, ', '.join(parts))
        else:
            msg = event_type

        return super()._log(level, msg, *args)


# Bootstrap the logger
logging.setLoggerClass(Logger)
getLogger = logging.getLogger



"""
Relevance, Full-stack search engine, core module.

Relevance is a software suite that aims to address the common and recurring problem
of ingesting content, delivering it, searching it with relevant results, analyzing
its usage by the end-users, acting on the analytic data to provide the best results
possible.

Its primary goals are rapid deployment, extensibility and ease of use.
"""

# Package version
__version__ = '0.4.2'

# Author information
__author__ = 'Francis Lacroix'

# License information
__license__ = 'MIT'
__copyright__ = 'OverrideLogic 2016'



"""
This module implements the search backend for ElasticSearch instances.
"""

import copy
import typing
import requests
import datetime
import warnings
from collections import OrderedDict

from relevance import loggers
from relevance.query import Search
from relevance.query import LogicalOperator
from relevance.query import ComparisonOperator
from relevance.facet import TermFacet
from relevance.facet import HistogramFacet
from relevance.facet import DateFacet
from relevance.facet import IntervalFacet
from relevance.facet import RangeFacet
from relevance.facet import FacetNotDefinedError
from relevance.facet import FacetNotSupportedWarning
from relevance.engine import Engine
from relevance.engine import EngineError
from relevance.engine import MappingError
from relevance.engine import SearchError
from relevance.engine import IngestionError
from relevance.mapping import Mapping
from relevance.mapping import Field
from relevance.mapping import ObjectField
from relevance.document import Document
from relevance.document import Result
from relevance.document import ResultSet


# Logging
logger = loggers.getLogger('relevance.engine.elastic')


class ElasticEngine(Engine):
    """
    This is the engine implementation for ElasticSearch backends.
    """
    def __init__(self, *args, **kwargs):
        """
        .. init
        """
        super().__init__(*args, **kwargs)
        self.session = None

    def to_dsl(self, search: Search) -> dict:
        """
        Transform a search object into a query DSL.

        :param search: the input search object.
        :returns: the resulting ElasticSearch query DSL.
        """
        def build_query(query):
            """
            Build a query into a DSL.
            """
            result = []

            # Logical operator
            if query.logic == LogicalOperator.AND:
                operator = 'must'
            elif query.logic == LogicalOperator.OR:
                operator = 'should'
            elif query.logic == LogicalOperator.NOT:
                operator = 'must_not'
            else:
                raise TypeError('unknown operator {0}'.format(query.logic))

            # Query terms
            for x in query.terms:
                result.append({'match': {'_all': x}})

            # Query filters
            for field, op, value in query.facets:
                start, end = value, value
                part = ref = {}

                # Manual facet definition
                try:
                    facet_def = self.facets[field]

                    if facet_def.path is not None:
                        field = facet_def.path

                        path = []
                        for x in facet_def.path.split('.')[:-1]:
                            path.append(x)
                            ref.update({
                                'nested': {
                                    'path': '.'.join(path),
                                    'query': {},
                                },
                            })
                            ref = ref['nested']['query']
                    elif facet_def.field is not None:
                        field = facet_def.field

                    if isinstance(facet_def, HistogramFacet):
                        start, end = facet_def.get_range_for(value)

                # Auto facet definition
                except (KeyError, TypeError):
                    if not self.auto_filters:
                        raise FacetNotDefinedError(field)

                # Filter operator
                if value is None:
                    if op == ComparisonOperator.EQ:
                        ref.update({'missing': {'field': field}})
                    else:
                        ref.update({'exists': {'field': field}})
                elif op == ComparisonOperator.EQ and start == end:
                    ref.update({'term': {field: value}})
                elif op == ComparisonOperator.NEQ and start == end:
                    ref.update({'bool': {'must_not': {'term': {field: value}}}})
                elif op == ComparisonOperator.EQ and start != end:
                    ref.update({'range': {field: {'gte': start, 'lt': end}}})
                elif op == ComparisonOperator.NEQ and start != end:
                    ref.update({'range': {field: {'lt': start, 'gte': end}}})
                elif op == ComparisonOperator.GT:
                    ref.update({'range': {field: {'gt': start}}})
                elif op == ComparisonOperator.LT:
                    ref.update({'range': {field: {'lt': end}}})
                elif op == ComparisonOperator.GTE:
                    ref.update({'range': {field: {'gte': start}}})
                elif op == ComparisonOperator.LTE:
                    ref.update({'range': {field: {'lte': end}}})
                else:
                    raise TypeError('unknown operator {0}'.format(op))

                result.append(part)

            # Sub queries
            for x in query.queries:
                result.append(build_query(x))

            dsl = {'bool': {operator: result}}
            if operator == 'should':
                dsl['bool']['minimum_should_match'] = 1

            logger.debug('query to dsl', query=query, dsl=dsl)
            return dsl

        def build_facets(search):
            """
            Build a search object into a aggregations DSL.
            """
            lst = list(self.facets) if search.facets is None else search.facets
            aggs = {}

            for field in lst:
                part = {field: {}}
                ref = part[field]

                # Manual facet definition
                try:
                    facet_def = self.facets[field]

                    if facet_def.path is not None:
                        field = facet_def.path

                        path = []
                        for x in facet_def.path.split('.')[:-1]:
                            path.append(x)
                            ref.update({
                                'nested': {
                                    'path': '.'.join(path),
                                },
                                'aggs': {'_nested': {}},
                            })
                            ref = ref['aggs']['_nested']
                    elif facet_def.field is not None:
                        field = facet_def.field

                # Auto facet definition
                except (KeyError, TypeError):
                    if not self.auto_filters:
                        raise FacetNotDefinedError(field)

                    facet_def = TermFacet(field)

                if isinstance(facet_def, TermFacet):
                    ref.update({
                        'terms': {'field': field}
                    })

                elif isinstance(facet_def, DateFacet):
                    ref.update({
                        'date_histogram': {
                            'field': field,
                            'interval': facet_def.interval,
                        }
                    })

                elif isinstance(facet_def, IntervalFacet):
                    ref.update({
                        'histogram': {
                            'field': field,
                            'interval': facet_def.interval,
                        }
                    })

                elif isinstance(facet_def, RangeFacet):
                    ref.update({
                        'range': {
                            'field': field,
                            'keyed': True,
                            'ranges': [{'key': k, 'from': v[0], 'to': v[1]}
                                       for k, v in facet_def.ranges.items()],
                        },
                    })

                else:
                    logger.warn('facet type not supported', type=facet_def.__class__)
                    warnings.warn(FacetNotSupportedWarning(facet_def))

                aggs.update(part)

            if len(aggs) > 0:
                logger.debug('facets to dsl', facets=lst, dsl=aggs)
                return aggs

        dsl = {}

        if search.queries is not None:
            dsl.update({
                'query': build_query(search.queries),
            })

        if search.slices:
            dsl.update({
                'from': search.slices[0],
                'size': search.slices[1],
            })

        if search.sorts:
            dsl.update({
                'sort': dict([(k, v.value) for k, v in search.sorts]),
            })

        if search.facets is None or len(search.facets) > 0:
            aggs = build_facets(search)
            if aggs is not None:
                dsl.update({
                    'aggs': aggs,
                })

        logger.debug('search to dsl', search=search, dsl=dsl)
        return dsl

    def search(self, search: Search) -> ResultSet:
        """
        Implementation of :meth:`.engine.Engine.search`.
        """
        if self.session is None:
            raise IOError('engine not started')

        super().search(search)

        url = '{0}/{1}_search'.format(
            self.target,
            ','.join(search.doc_types) + '/' if search.doc_types is not None else '',
        )
        payload = self.to_dsl(search)

        logger.debug('http POST', url=url, length=len(payload))

        try:
            results = ResultSet()
            response = self.session.post(url, json=payload)
            response.raise_for_status()
            response = response.json()
            results.time_end = datetime.datetime.now()
            results.count = response['hits']['total']
        except self.session.exceptions.HTTPError as e:
            raise SearchError(str(e))
        except Exception as e:
            raise EngineError(str(e))

        logger.debug('http response', response=response)

        for x in response['hits']['hits']:
            result = Result(x['_index'], x['_type'], x['_id'], score=x['_score'])
            result.update(x['_source'])
            results.append(result)

        if 'aggregations' in response:
            results.facets = {}
            for name, data in response['aggregations'].items():
                results.facets[name] = OrderedDict()

                ref = data
                while '_nested' in ref:
                    ref = ref['_nested']

                if 'buckets' not in ref:
                    continue

                if isinstance(ref['buckets'], dict):
                    for x, y in ref['buckets'].items():
                        results.facets[name][x] = y['doc_count']
                elif isinstance(ref['buckets'], list):
                    for x in ref['buckets']:
                        key = x.get('key_as_string', x.get('key'))
                        results.facets[name][key] = x['doc_count']

        return results

    def get_doc_types(self) -> typing.List[str]:
        """
        Implementation of :meth:`.engine.Engine.get_doc_types`.
        """
        if self.session is None:
            raise IOError('engine not started')

        url = '{0}/_mapping'.format(self.target)

        logger.debug('http GET', url=url)

        try:
            response = self.session.get(url)
            response.raise_for_status()
            response = response.json()
        except self.session.exceptions.HTTPError as e:
            raise MappingError(str(e))
        except Exception as e:
            raise EngineError(str(e))

        logger.debug('http response', response=response)

        result = set()
        for index in response:
            try:
                for doc_type in response[index]['mappings']:
                    result.add(doc_type)
            except KeyError:
                pass

        return list(result)

    def get_mapping(self, doc_type: str) -> Mapping:
        """
        Implementation of :meth:`.engine.Engine.get_mapping`.
        """
        if self.session is None:
            raise IOError('engine not started')

        url = '{0}/_mapping'.format(self.target)

        logger.debug('http GET', url=url)

        try:
            response = self.session.get(url)
            response.raise_for_status()
            response = response.json()
        except self.session.exceptions.HTTPError as e:
            raise MappingError(str(e))
        except Exception as e:
            raise EngineError(str(e))

        logger.debug('http response', response=response)

        for index in response:
            def build_mapping(properties):
                """
                Build a mapping object from a properties dict.
                """
                mapping = Mapping()

                for key, conf in properties.items():
                    options = dict(
                        stored=conf.get('store', False),
                        indexed=conf.get('index', False),
                    )

                    if conf['type'] == 'string':
                        field = Field(key, str, **options)
                    if conf['type'] in ('long', 'integer'):
                        field = Field(key, int, **options)
                    if conf['type'] == 'double':
                        field = Field(key, float, **options)
                    if conf['type'] == 'boolean':
                        field = Field(key, bool, **options)
                    if conf['type'] == 'object':
                        field = Field(key, dict, **options)
                    if conf['type'] == 'binary':
                        field = Field(key, bytes, **options)
                    if conf['type'] == 'date':
                        field = Field(key, datetime.datetime, **options)
                    if conf['type'] == 'nested':
                        field = ObjectField(key, build_mapping(conf['properties']))

                    mapping.add(field)

                return mapping

            try:
                properties = response[index]['mappings'][doc_type]['properties']
            except KeyError:
                continue

            return build_mapping(properties)

        raise MappingError('no mapping for {0}'.format(doc_type))

    def put_mapping(self, doc_type: str, mapping: Mapping):
        """
        Implementation of :meth:`.engine.Engine.put_mapping`.
        """
        if self.session is None:
            raise IOError('engine not started')

        raise NotImplementedError()

    def publish(self, doc: Document) -> Document:
        """
        Implementation of :meth:`.engine.Engine.publish`.
        """
        if self.session is None:
            raise IOError('engine not started')

        url = '{0}/{1}/{2}'.format(self.target, doc.schema, doc.doc_type)

        payload = dict(doc)
        if doc.uid is not None:
            payload['_id'] = doc.uid

        logger.debug('http POST', url=url, length=len(payload))

        try:
            response = self.session.post(url, json=payload)
            response.raise_for_status()
            response = response.json()
        except self.session.exceptions.HTTPError as e:
            raise IngestionError(str(e))
        except Exception as e:
            raise EngineError(str(e))

        logger.debug('http response', response=response)

        doc = copy.deepcopy(doc)
        doc.uid = response.get('_id', doc.uid)
        return doc

    def unpublish(self, schema: str, doc_type: str, uid: object):
        """
        Implementation of :meth:`.engine.Engine.unpublish`.
        """
        if self.session is None:
            raise IOError('engine not started')

        url = '{0}/{1}/{2}/{3}'.format(self.target, schema, doc_type, uid)

        logger.debug('http DELETE', url=url)

        try:
            response = self.session.delete(url)
            response.raise_for_status()
        except self.session.exceptions.HTTPError as e:
            raise IngestionError(str(e))
        except Exception as e:
            raise EngineError(str(e))

        logger.debug('http response', response=response)

    def start(self):
        """
        Implementation of :meth:`.engine.Engine.start`.
        """
        if self.session is not None:
            return

        logger.info('start', id=id(self.session))

        self.session = requests.Session()
        self.session.get(self.target)

    def stop(self):
        """
        Implementation of :meth:`.engine.Engine.stop`.
        """
        if self.session is None:
            return

        logger.info('stop', id=id(self.session))

        self.session.close()
        self.session = None

    def get_status(self) -> typing.Dict[str, object]:
        """
        Get the indexer state.

        :returns: a dictionary containing various information about the status.
        """
        if self.session is None:
            raise IOError('engine not started')

        url = '{0}/_stats'.format(self.target)

        logger.debug('http GET', url=url)

        try:
            response = self.session.get(url)
            response.raise_for_status()
            response = response.json()['_all']['total']
        except Exception as e:
            raise EngineError(str(e))

        logger.debug('http response', response=response)

        return {
            'indexed': response['indexing']['index_total'],
            'deleted': response['indexing']['delete_total'],
            'commits': response['refresh']['total'],
            'size': response['store']['size_in_bytes'],
        }



"""
This module provides base classes for services to expose functionality via APIs.
"""

import flask
import typing
from argparse import Namespace

from relevance.settings import SettingsFactory


class Service(flask.Flask):
    """
    This class wraps the :class:`flask.Flask` application object and provides additional
    features to it, such as common error handling and response formatting.
    """
    instances = {}
    """ The service instances mapping. """

    def __init__(self, name: str, *args, **kwargs):
        """
        :param name: the name of the service.
        """
        self.svc_name = name
        self.data = Namespace()
        imp_name = 'relevance.{0}.service'.format(name)
        super().__init__(imp_name, *args, **kwargs)

        self.before_start_funcs = []
        self.before_start_done = False

        # Initialize settings
        loader = SettingsFactory(['/etc/relevance', './etc'])
        self.settings = loader.load(self.svc_name)

        @self.errorhandler(405)
        def error_405(_):
            return self.result(405, {
                'error': {
                    'type': 'MethodNotAllowedError',
                    'desc': 'the requested method is not allowed for the requested resource',
                    'resource': self.request.path,
                    'key': self.request.method,
                }
            })

    @classmethod
    def instance(cls, name: str, *args, **kwargs) -> 'Service':
        """
        Create or get an instance of a service.

        In some environments like multiprocessing, importing a global service object has
        not the intended effect. This method ensures that the instance that is retrieved
        is always the same across threads.

        If an instance `name` already exists, it will be returned. Otherwise, one will be
        created, stored and returned.

        :param name: the instance name.
        :returns: a service object.
        """
        if name in cls.instances:
            return cls.instances[name]
        instance = cls(name, *args, **kwargs)
        cls.instances[name] = instance
        return instance

    def before_start(self, func: typing.Callable) -> typing.Callable:
        """
        :decorator: Decorator method for functions to call before starting the application.
            Useful for loading configuration files, per example.

        :param func: the function to decorate.
        :returns: the decorated function.
        """
        self.before_start_funcs.append(func)
        return func

    @property
    def request(self) -> flask.request:
        """
        Get the current request.

        :returns: the Flask request object.
        """
        return flask.request

    def result(self, status_code: int, content: object=None) -> flask.Response:
        """
        Obtain a response object.

        :param status_code: the status code to return.
        :param content: the content to return. If the content is a dictionary, a list
        or a tuple. It will be formatted to JSON. Otherwise, it will be returned as is.
        :returns: a response object.
        """
        if content is None:
            response = flask.Response()
        elif isinstance(content, (dict, list, tuple)):
            response = flask.jsonify(content)
        else:
            response = flask.Response()
            response.content = content
        response.status_code = status_code
        return response



"""
This package provides a unified interface for interacting with the different storage
and search backends.
"""

import abc
import pydoc
import typing

from relevance import loggers
from relevance.settings import Settings
from relevance.query import Search
from relevance.facet import Facet
from relevance.mapping import Mapping
from relevance.document import Document
from relevance.document import ResultSet


# Logging
logger = loggers.getLogger('relevance.engine')


class Engine(object, metaclass=abc.ABCMeta):
    """
    This is the interface for search engines. Implementing this interface allows
    a search engine to be queried and documents to be indexed to it.
    """
    def __init__(self, target: str, *, facets: typing.Dict[str, Facet]=None,
                 auto_filters: bool=True, **kwargs):
        """
        :param target: the storage engine target, typically a URL or DSN.
        :param facets: a dictionary of facets definitions. The key is the how the
            facet will be named when queried, the value is a :class:`.facet.Facet`
            object.
        :param auto_filters: whether to enable automatic filter definitions. When
            enabled, if a facet has not been defined, and one is specified in the
            search request, whether as a filter or in the returned facets list,
            it is assumed to be a :class:`.facet.TermFacet`. Even when enabled,
            undefined facets will not be returned by default, and will have to
            be specified manually in the search request. When disabled, specifying
            an undefined facet in a request will raise a
            :class:`.facet.FacetNotDefinedError`.
        """
        logger.info('init', inst=self, target=target, auto_filters=auto_filters,
                    facets=facets)
        self.target = target
        self.facets = facets if facets is not None else {}
        self.auto_filters = auto_filters

    @abc.abstractmethod
    def search(self, search: Search) -> ResultSet:
        """
        Perform a search request on the engine.

        An implementation of this method should accept a search object, and return
        a result set object on completion.

        :param search: the request to execute.
        :returns: a result set object.
        """
        pass

    @abc.abstractmethod
    def get_doc_types(self) -> typing.List[str]:
        """
        Get a list of available document types.

        Most engines can support multiple "document types". For more database-like
        engines, this could be a table name or an object type.

        :returns: a list of document type strings supported by the engine.
        """
        pass

    @abc.abstractmethod
    def get_mapping(self, doc_type: str) -> Mapping:
        """
        Get the mapping object for a specific document type.

        Given a specific document type as argument, this method should construct a
        :class:`.mapping.Mapping` object from :class:`.mapping.Field` objects and
        return it based on the mapping in the search backend.

        :param doc_type: the document type.
        :returns: the mapping object.
        """
        pass

    @abc.abstractmethod
    def put_mapping(self, doc_type: str, mapping: Mapping):
        """
        Update the mapping for a specific document type.

        Calling this method should update the mapping on the search backend.

        :param doc_type: the document type.
        :param mapping: the mapping object.
        """
        pass

    @abc.abstractmethod
    def publish(self, doc: Document) -> Document:
        """
        Index a document.

        This method should return a copy of the original document, with auto fields
        (e.g.: id) populated automatically.

        :param doc: the document object to index.
        :returns: the updated, indexed document object.
        """
        pass

    @abc.abstractmethod
    def unpublish(self, schema: str, doc_type: str, uid: object):
        """
        Delete a document.

        :param schema: the document schema.
        :param doc_type: the document type.
        :param uid: the unique identifier for the document.
        """
        pass

    def start(self):
        """
        Start the engine.

        Some backends are not stateless and require setup and teardown. The setup part
        can be implemented here.
        """
        pass

    def stop(self):
        """
        Stop the engine.

        Some backends are not stateless and require setup and teardown. The teardown part
        can be implemented here.
        """
        pass


class EngineFactory(object):
    """
    This factory class allows to create :class:`Engine` objects from :class:`.settings.Settings`
    objects.
    """
    def load_facet(self, settings: Settings) -> Facet:
        """
        Obtain a facet object from settings.

        :param settings: the settings object to load.
        :returns: the facet object.
        :raises: :class:`ImportError`: if the facet class is invalid.
        """
        cls_name = settings.get_type('class', data_type=str)
        field = settings.get('field', data_type=str)
        path = settings.get('path', data_type=str)
        options = settings.get('options', {}, data_type=dict)

        # locate the facet class
        cls = pydoc.locate(cls_name)
        if cls is None or not issubclass(cls, Facet):
            raise ImportError(cls_name)

        result = cls(field=field, path=path, **options)
        return result

    def load_all_facets(self, settings: Settings) -> typing.Dict[str, Facet]:
        """
        Obtain a mapping of facets from settings.

        :param settings: the settings object to load.
        :returns a dicionary of facets. It can then be passed directly to the
            :class:`Engine` initializer.
        """
        result = {}
        for name, conf in settings.items():
            result[name] = self.load_facet(conf)
        return result

    def load_engine(self, settings: Settings) -> Engine:
        """
        Obtain an engine object from settings.

        :param settings: the settings object to load.
        :returns: the engine object.
        :raises: :class:`ImportError`: if the engine class is invalid.
        """
        cls_name = settings.get_type('class', data_type=str)
        target = settings.get_type('target', data_type=str)
        options = settings.get('options', {}, data_type=dict)
        facets = settings.get('facets', {}, data_type=dict)

        # locate the engine class
        cls = pydoc.locate(cls_name)
        if cls is None or not issubclass(cls, Engine):
            raise ImportError(cls_name)

        # load the facets
        facets_map = self.load_all_facets(facets)

        result = cls(target, facets=facets_map, **options)
        return result

    def load_all_engines(self, settings: Settings) -> typing.Dict[str, Engine]:
        """
        Obtain a mapping of engine objects.

        :param settings: the settings object to load.
        :returns: a dictionary object mapping the engine objects to names. This can
            then be fed into a :class:`.manager.ResourceManager` object.
        """
        result = {}
        for name, conf in settings.items():
            result[name] = self.load_engine(conf)
        return result


class EngineError(RuntimeError):
    """
    This is the base exception class when an engine encounters an error.
    """
    pass


class SearchError(EngineError):
    """
    This exception type is raised when an error occurs trying to execute a search
    request.
    """
    pass


class MappingError(EngineError):
    """
    This exception type is raised when trying to get or update mapping in a
    search backend and an error occurs.
    """
    pass


class IngestionError(EngineError):
    """
    This exception type is raised when an error occurs trying to publish or unpublish
    documents in a search backend.
    """
    pass



"""
This pipeline adapter module exposes queues to be used as extractors or publishers.
They can be populated or retrieved via service routes.
"""

import pickle
from queue import Queue
from queue import Empty
from queue import Full
from werkzeug.exceptions import NotFound

from relevance.pipeline import Extractor
from relevance.pipeline import Publisher
from relevance.pipeline import Pipeline
from relevance.worker import Worker
from relevance.service import Service
from relevance.manager import ResourceNotFoundError


# Get the service instance
service = Service.instance('pipeline')


class QueueExtractor(Extractor):
    """
    This class exposes a queue object within a worker that can be used as an
    extractor. Items can be put in the queue via the service route.
    """
    def __init__(self, worker: Worker, size: int=0, timeout: float=0):
        """
        :param worker: the worker object for the extractor.
        :param size: the maximum size of the queue. Once the queue has been
            filled, further put calls will hang until there is room or the
            timeout is expired.
        :param timeout: a timeout, in seconds, after which requests to and
            from the queue expire and raise an exception.
        """
        self.queue = Queue(maxsize=size)
        self.timeout = timeout
        super().__init__(worker, self.queue.__class__)

    def __call__(self, pipeline: Pipeline) -> object:
        """
        Implementation of :meth:`Extractor.__call__`.
        """
        try:
            item = self.queue.get(timeout=self.timeout)
            self.queue.task_done()
        except Empty:
            return StopIteration()

        pipeline.trigger('item_extracted', item=item)
        return item

    def put(self, item: object):
        """
        Put an item in the queue.

        :param item: the item to put.
        :raises: :class:`queue.Full`: if the timeout specified in the initializer
            is reached when the queue is full.
        """
        self.queue.put(item, timeout=self.timeout)


@service.route('/pipelines/<pipeline_name>/components/<component_name>/queue',
               methods=['POST'])
def put_pipeline_queue_item(pipeline_name: str, component_name: str):
    """
    Put an item into an input queue.

    :Method: POST
    :Path: /pipelines/`pipeline_name`/components/`component_name`/queue
    :Headers: requires a proper `Content-Type` header to be set.
    :param pipeline_name: the name of the pipeline.
    :param pipeline_name: the name of the component.

    :Statuses: 201 on success.
    :Response:
        .. code-block:: json

            {
                "content_type": "application/json",
                "pending_tasks": 42,
                "size": 129
            }
    :raises: :class:`UnsupportedContentTypeError`: if the `Content-Type` header does
        not contain a supported mime type.
    """
    manager = service.data.manager

    try:
        pipeline = manager.resolve(pipeline_name)
    except KeyError as e:
        raise ResourceNotFoundError('pipeline', eval(str(e)))

    try:
        component = pipeline.get(component_name)
    except KeyError as e:
        raise ResourceNotFoundError('component', eval(str(e)))

    if not isinstance(component, QueueExtractor):
        raise NotFound()

    content_type = service.request.headers.get('Content-Type', 'text/plain')

    if content_type == 'text/plain':
        item = service.request.get_data()
    elif content_type == 'application/json':
        item = service.request.get_json()
    elif content_type == 'application/octet-stream':
        item = service.request.get_data()
        item = pickle.loads(item)
    else:
        raise UnsupportedContentTypeError(content_type)

    component.put(item)

    result = {
        'content_type': content_type,
        'pending_tasks': component.queue.qsize(),
        'size': len(pickle.dumps(item)),
    }
    return service.result(201, result)


class QueuePublisher(Publisher):
    """
    The queue publisher is a pipeline component implementation that allows a document to
    be published to an output queue.
    """
    def __init__(self, worker: Worker, size: int=0, timeout: float=0):
        """
        :param worker: the worker object for the publisher.
        :param size: the maximum size of the queue. Once the queue has been
            filled, further calls will hang until there is room or the
            timeout is expired.
        :param timeout: a timeout, in seconds, after which requests to and
            from the queue expire and raise an exception.
        """
        self.queue = Queue(maxsize=size)
        self.timeout = timeout
        super().__init__(worker, self.queue.__class__)

    def __call__(self, pipeline: Pipeline, item: object) -> object:
        """
        Implementation of :meth:`Extractor.__call__`.
        """
        try:
            self.queue.put(item, timeout=self.timeout)
        except Full:
            return StopIteration()

        pipeline.trigger('item_published', item=item)
        return item

    def get(self) -> object:
        """
        Get an item from the queue.

        :returns: the retrieved item.
        :raises: :class:`queue.Empty`: if the queue is empty and the timeout is
            reached.
        """
        return self.queue.get(timeout=self.timeout)


@service.route('/pipelines/<pipeline_name>/components/<component_name>/queue',
               methods=['GET'])
def get_pipeline_queue_item(pipeline_name: str, component_name: str):
    """
    Get an item from an output queue.

    :Method: GET
    :Path: /pipelines/`pipeline_name`/components/`component_name`/queue
    :Headers: requires a proper `Accept` header to be set.
    :param pipeline_name: the name of the pipeline.
    :param pipeline_name: the name of the component.

    :Statuses: 200 on success, 204 if the queue is empty.
    :Response: the raw document in the format specified by the `Accept` header.
    :raises: :class:`UnsupportedContentTypeError`: if the `Accept` header does
        not contain a supported mime type.
    """
    manager = service.data.manager

    try:
        pipeline = manager.resolve(pipeline_name)
    except KeyError as e:
        raise ResourceNotFoundError('pipeline', eval(str(e)))

    try:
        component = pipeline.get(component_name)
    except KeyError as e:
        raise ResourceNotFoundError('component', eval(str(e)))

    if not isinstance(component, QueuePublisher):
        raise NotFound()

    content_type = service.request.headers.get('Accept', 'text/plain')

    if content_type == 'text/plain':
        parser = None
    elif content_type == 'application/json':
        parser = None
    elif content_type == 'application/octet-stream':
        parser = pickle.dumps
    else:
        raise UnsupportedContentTypeError(content_type)

    try:
        item = component.get()
    except Empty:
        return service.result(204)

    if parser is not None:
        item = parser(item)

    return service.result(200, item)


class UnsupportedContentTypeError(TypeError):
    """
    This exception is raised when the specified mime type in the `Content-Type` or
    `Accept` header of a queue component request is not supported.
    """
    pass


@service.errorhandler(UnsupportedContentTypeError)
def _error_unsupported_content_type(e):
    """
    Handle :class:`UnsupportedContentTypeError` errors.
    """
    return service.result(415, {
        'error': {
            'type': e.__class__.__name__,
            'desc': 'the content type specified in the requested is not supported',
            'content_type': str(e),
        }
    })



#!/usr/bin/python3

from setuptools import setup
from setuptools import find_packages

import relevance as main


setup(
    name=main.__name__,
    version=main.__version__,

    description=main.__doc__.split('\n')[1].strip(),
    long_description=main.__doc__.strip(),
    url='http://www.relevance.io',
    author='OverrideLogic',
    author_email='info@overridelogic.com',
    maintainer='Francis Lacroix',
    maintainer_email='f@overridelogic.com',

    license='MIT',
    platforms=['any'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search'
    ],

    packages=find_packages(exclude=['tests', 'tests.*']),
    provides=[main.__name__],

    python_requires='>=3.5',
    setup_requires=['flake8>=3.2.1'],
    tests_require=[],
    test_suite='tests',

    install_requires=[
        'Werkzeug>=0.12.2',
        'Flask>=0.12',
        'anyconfig>=0.9.1',
        'Twisted>=17.9.0',
        'python-dateutil>=2.6.1',
        'requests>=2.9.1',
    ],

    entry_points={
        'console_scripts': [
            'relevance=relevance.utils.cli:main',
            'relevance-server=relevance.utils.cli:server',
            'relevance-search=relevance.utils.cli:search',
        ],
    },

    data_files=[
        ('/etc/relevance', [
            'etc/search.json-example',
        ]),
    ],
)
